# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'tikka/slots/pyqt/resources/gui/windows/indexer_add.ui'
#
# Created by: PyQt5 UI code generator 5.15.10
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_indexerAddDialog(object):
    def setupUi(self, indexerAddDialog):
        indexerAddDialog.setObjectName("indexerAddDialog")
        indexerAddDialog.resize(731, 126)
        self.verticalLayout = QtWidgets.QVBoxLayout(indexerAddDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.urlLabel = QtWidgets.QLabel(indexerAddDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.urlLabel.setFont(font)
        self.urlLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.urlLabel.setObjectName("urlLabel")
        self.gridLayout.addWidget(self.urlLabel, 0, 0, 1, 1)
        self.blockValueLabel = QtWidgets.QLabel(indexerAddDialog)
        self.blockValueLabel.setText("")
        self.blockValueLabel.setObjectName("blockValueLabel")
        self.gridLayout.addWidget(self.blockValueLabel, 1, 1, 1, 1)
        self.urlLineEdit = QtWidgets.QLineEdit(indexerAddDialog)
        self.urlLineEdit.setObjectName("urlLineEdit")
        self.gridLayout.addWidget(self.urlLineEdit, 0, 1, 1, 1)
        self.testButton = QtWidgets.QPushButton(indexerAddDialog)
        self.testButton.setObjectName("testButton")
        self.gridLayout.addWidget(self.testButton, 0, 2, 1, 1)
        self.blockLabel = QtWidgets.QLabel(indexerAddDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.blockLabel.setFont(font)
        self.blockLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.blockLabel.setObjectName("blockLabel")
        self.gridLayout.addWidget(self.blockLabel, 1, 0, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout)
        self.errorLabel = QtWidgets.QLabel(indexerAddDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.errorLabel.sizePolicy().hasHeightForWidth())
        self.errorLabel.setSizePolicy(sizePolicy)
        self.errorLabel.setStyleSheet("color: red")
        self.errorLabel.setText("")
        self.errorLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.errorLabel.setObjectName("errorLabel")
        self.verticalLayout.addWidget(self.errorLabel)
        self.buttonBox = QtWidgets.QDialogButtonBox(indexerAddDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setCenterButtons(True)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(indexerAddDialog)
        self.buttonBox.accepted.connect(indexerAddDialog.accept) # type: ignore
        self.buttonBox.rejected.connect(indexerAddDialog.reject) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(indexerAddDialog)
        indexerAddDialog.setTabOrder(self.urlLineEdit, self.testButton)

    def retranslateUi(self, indexerAddDialog):
        #  _translate = QtCore.QCoreApplication.translate
        indexerAddDialog.setWindowTitle(_("Add indexer"))
        self.urlLabel.setText(_("URL"))
        self.testButton.setText(_("Test"))
        self.blockLabel.setText(_("Block"))
