# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'tikka/slots/pyqt/resources/gui/windows/category_add.ui'
#
# Created by: PyQt5 UI code generator 5.15.10
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_addCategoryDialog(object):
    def setupUi(self, addCategoryDialog):
        addCategoryDialog.setObjectName("addCategoryDialog")
        addCategoryDialog.resize(400, 111)
        self.gridLayout = QtWidgets.QGridLayout(addCategoryDialog)
        self.gridLayout.setObjectName("gridLayout")
        self.buttonBox = QtWidgets.QDialogButtonBox(addCategoryDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.gridLayout.addWidget(self.buttonBox, 2, 0, 1, 2, QtCore.Qt.AlignHCenter)
        self.nameLabel = QtWidgets.QLabel(addCategoryDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.nameLabel.setFont(font)
        self.nameLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.nameLabel.setObjectName("nameLabel")
        self.gridLayout.addWidget(self.nameLabel, 0, 0, 1, 1)
        self.nameValueLineEdit = QtWidgets.QLineEdit(addCategoryDialog)
        self.nameValueLineEdit.setObjectName("nameValueLineEdit")
        self.gridLayout.addWidget(self.nameValueLineEdit, 0, 1, 1, 1)
        self.parentComboBox = QtWidgets.QComboBox(addCategoryDialog)
        self.parentComboBox.setObjectName("parentComboBox")
        self.gridLayout.addWidget(self.parentComboBox, 1, 1, 1, 1)
        self.parentLabel = QtWidgets.QLabel(addCategoryDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.parentLabel.setFont(font)
        self.parentLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.parentLabel.setObjectName("parentLabel")
        self.gridLayout.addWidget(self.parentLabel, 1, 0, 1, 1)

        self.retranslateUi(addCategoryDialog)
        self.buttonBox.accepted.connect(addCategoryDialog.accept) # type: ignore
        self.buttonBox.rejected.connect(addCategoryDialog.reject) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(addCategoryDialog)

    def retranslateUi(self, addCategoryDialog):
        #  _translate = QtCore.QCoreApplication.translate
        addCategoryDialog.setWindowTitle(_("Add category"))
        self.nameLabel.setText(_("Name"))
        self.parentLabel.setText(_("Parent"))
