# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'tikka/slots/pyqt/resources/gui/widgets/account_tree.ui'
#
# Created by: PyQt5 UI code generator 5.15.10
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_AccountTreeWidget(object):
    def setupUi(self, AccountTreeWidget):
        AccountTreeWidget.setObjectName("AccountTreeWidget")
        AccountTreeWidget.resize(664, 447)
        self.verticalLayout = QtWidgets.QVBoxLayout(AccountTreeWidget)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.addCategoryButton = QtWidgets.QPushButton(AccountTreeWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.addCategoryButton.sizePolicy().hasHeightForWidth())
        self.addCategoryButton.setSizePolicy(sizePolicy)
        self.addCategoryButton.setObjectName("addCategoryButton")
        self.horizontalLayout.addWidget(self.addCategoryButton)
        self.refreshButton = QtWidgets.QPushButton(AccountTreeWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.refreshButton.sizePolicy().hasHeightForWidth())
        self.refreshButton.setSizePolicy(sizePolicy)
        self.refreshButton.setObjectName("refreshButton")
        self.horizontalLayout.addWidget(self.refreshButton, 0, QtCore.Qt.AlignRight)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.errorLabel = QtWidgets.QLabel(AccountTreeWidget)
        self.errorLabel.setStyleSheet("color: red")
        self.errorLabel.setText("")
        self.errorLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.errorLabel.setObjectName("errorLabel")
        self.horizontalLayout_2.addWidget(self.errorLabel)
        self.totalBalanceValueLabel = QtWidgets.QLabel(AccountTreeWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.totalBalanceValueLabel.sizePolicy().hasHeightForWidth())
        self.totalBalanceValueLabel.setSizePolicy(sizePolicy)
        self.totalBalanceValueLabel.setStyleSheet("color: white;\n"
"border : 3px solid rgb(34, 156, 244);\n"
"border-radius: 10%;\n"
"background: rgb(34, 156, 244);\n"
"\n"
"\n"
"")
        self.totalBalanceValueLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.totalBalanceValueLabel.setObjectName("totalBalanceValueLabel")
        self.horizontalLayout_2.addWidget(self.totalBalanceValueLabel)
        self.verticalLayout.addLayout(self.horizontalLayout_2)
        self.treeView = QtWidgets.QTreeView(AccountTreeWidget)
        self.treeView.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.treeView.setSizeAdjustPolicy(QtWidgets.QAbstractScrollArea.AdjustToContents)
        self.treeView.setDragEnabled(True)
        self.treeView.setDragDropOverwriteMode(True)
        self.treeView.setDragDropMode(QtWidgets.QAbstractItemView.InternalMove)
        self.treeView.setAlternatingRowColors(True)
        self.treeView.setIndentation(20)
        self.treeView.setItemsExpandable(True)
        self.treeView.setHeaderHidden(True)
        self.treeView.setExpandsOnDoubleClick(True)
        self.treeView.setObjectName("treeView")
        self.verticalLayout.addWidget(self.treeView)

        self.retranslateUi(AccountTreeWidget)
        QtCore.QMetaObject.connectSlotsByName(AccountTreeWidget)
        AccountTreeWidget.setTabOrder(self.addCategoryButton, self.treeView)

    def retranslateUi(self, AccountTreeWidget):
        #  _translate = QtCore.QCoreApplication.translate
        AccountTreeWidget.setWindowTitle(_("Accounts"))
        self.addCategoryButton.setText(_("Add category"))
        self.refreshButton.setText(_("Refresh"))
        self.totalBalanceValueLabel.setText(_("?"))
import tikka.slots.pyqt.resources.icons.index_rc
