# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'tikka/slots/pyqt/resources/gui/widgets/servers.ui'
#
# Created by: PyQt5 UI code generator 5.15.10
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_ServersWidget(object):
    def setupUi(self, ServersWidget):
        ServersWidget.setObjectName("ServersWidget")
        ServersWidget.resize(724, 526)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(ServersWidget)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.nodesGroupBox = QtWidgets.QGroupBox(ServersWidget)
        self.nodesGroupBox.setObjectName("nodesGroupBox")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.nodesGroupBox)
        self.verticalLayout.setObjectName("verticalLayout")
        self.addNodeButton = QtWidgets.QPushButton(self.nodesGroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.addNodeButton.sizePolicy().hasHeightForWidth())
        self.addNodeButton.setSizePolicy(sizePolicy)
        self.addNodeButton.setObjectName("addNodeButton")
        self.verticalLayout.addWidget(self.addNodeButton, 0, QtCore.Qt.AlignRight)
        self.nodesTableView = QtWidgets.QTableView(self.nodesGroupBox)
        self.nodesTableView.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.nodesTableView.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.nodesTableView.setTabKeyNavigation(False)
        self.nodesTableView.setProperty("showDropIndicator", False)
        self.nodesTableView.setAlternatingRowColors(True)
        self.nodesTableView.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.nodesTableView.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.nodesTableView.setShowGrid(False)
        self.nodesTableView.setSortingEnabled(True)
        self.nodesTableView.setObjectName("nodesTableView")
        self.nodesTableView.horizontalHeader().setStretchLastSection(True)
        self.nodesTableView.verticalHeader().setVisible(False)
        self.verticalLayout.addWidget(self.nodesTableView)
        self.verticalLayout_2.addWidget(self.nodesGroupBox)
        self.indexersGroupBox = QtWidgets.QGroupBox(ServersWidget)
        self.indexersGroupBox.setObjectName("indexersGroupBox")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.indexersGroupBox)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.addIndexerButton = QtWidgets.QPushButton(self.indexersGroupBox)
        self.addIndexerButton.setObjectName("addIndexerButton")
        self.verticalLayout_3.addWidget(self.addIndexerButton, 0, QtCore.Qt.AlignRight)
        self.indexersTableView = QtWidgets.QTableView(self.indexersGroupBox)
        self.indexersTableView.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.indexersTableView.setTabKeyNavigation(False)
        self.indexersTableView.setProperty("showDropIndicator", False)
        self.indexersTableView.setAlternatingRowColors(True)
        self.indexersTableView.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.indexersTableView.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.indexersTableView.setShowGrid(False)
        self.indexersTableView.setSortingEnabled(True)
        self.indexersTableView.setObjectName("indexersTableView")
        self.indexersTableView.horizontalHeader().setStretchLastSection(True)
        self.indexersTableView.verticalHeader().setVisible(False)
        self.verticalLayout_3.addWidget(self.indexersTableView)
        self.verticalLayout_2.addWidget(self.indexersGroupBox)

        self.retranslateUi(ServersWidget)
        QtCore.QMetaObject.connectSlotsByName(ServersWidget)

    def retranslateUi(self, ServersWidget):
        #  _translate = QtCore.QCoreApplication.translate
        ServersWidget.setWindowTitle(_("Network servers"))
        self.nodesGroupBox.setTitle(_("Nodes"))
        self.addNodeButton.setText(_("Add node"))
        self.indexersGroupBox.setTitle(_("Indexers"))
        self.addIndexerButton.setText(_("Add indexer"))
import tikka.slots.pyqt.resources.icons.index_rc
